import shutil
from pathlib import Path

import click
from ape import project

PLUGIN_MANIFEST_FOLDER = project.path / "ape_safe" / "manifests"


@click.command()
def cli():
    for version in project.dependencies["safe-contracts"]:
        dependency = project.dependencies["safe-contracts"][version]
        plugin_manifest_path = PLUGIN_MANIFEST_FOLDER / f"safe-{version}.json"
        click.echo(
            f"cp $HOME/{dependency.manifest_path.relative_to(Path.home())}"
            f" ./{plugin_manifest_path.relative_to(Path.cwd())}"
        )
        shutil.copyfile(dependency.manifest_path, plugin_manifest_path)



import json
import tempfile
from pathlib import Path

import pytest
from ape.contracts import ContractContainer
from ethpm_types import ContractType
from packaging.version import Version

from ape_safe.accounts import SafeAccount
from ape_safe.factory import SafeFactory

contracts_directory = Path(__file__).parent / "contracts"


@pytest.fixture(
    scope="session",
    # TODO: Test more versions.
    params=(
        "1.4.1",
        "1.3.0",
        "1.1.1",
    ),
)
def VERSION(request):
    return Version(request.param)


@pytest.fixture(scope="session")
def deployer(OWNERS):
    return OWNERS[-1]


@pytest.fixture(scope="session")
def safe_factory(VERSION, deployer):
    SafeFactory.inject(VERSION, deployer)
    return SafeFactory()


@pytest.fixture(scope="session")
def receiver(accounts):
    return accounts[9]


@pytest.fixture(scope="session")
def delegate(accounts):
    return accounts[8]


@pytest.fixture(scope="session", params=["1/1", "1/2", "2/2", "2/3", "3/3"])
def MULTISIG_TYPE(request):
    # Param is `M/N`, but encoded as a string for repr in pytest
    return request.param.split("/")


@pytest.fixture(scope="session")
def THRESHOLD(MULTISIG_TYPE):
    M, _ = MULTISIG_TYPE
    return int(M)


@pytest.fixture(scope="session")
def OWNERS(accounts, MULTISIG_TYPE):
    _, N = MULTISIG_TYPE
    return accounts[: int(N)]


@pytest.fixture
def safe_contract(safe_factory, deployer, OWNERS, THRESHOLD):
    return safe_factory.create(OWNERS, THRESHOLD, sender=deployer)


@pytest.fixture
def safe_data_file(chain, safe_contract):
    with tempfile.NamedTemporaryFile() as fp:
        file = Path(str(fp.name))
        file.write_text(
            json.dumps(
                {
                    "address": safe_contract.address,
                    "deployed_chain_ids": [chain.provider.chain_id],
                }
            )
        )
        yield file


@pytest.fixture
def safe(safe_data_file):
    return SafeAccount(account_file_path=safe_data_file)


@pytest.fixture
def token(deployer: SafeAccount):
    text = (contracts_directory / "Token.json").read_text()
    contract = ContractType.model_validate_json(text)
    return deployer.deploy(ContractContainer(contract))


@pytest.fixture
def vault(deployer: SafeAccount, token):
    text = (contracts_directory / "VyperVault.json").read_text()
    vault = ContractContainer(ContractType.model_validate_json(text))
    return deployer.deploy(vault, token)



from datetime import datetime, timezone
from enum import Enum
from typing import Annotated, NewType, Optional, Union, cast

from ape.types import AddressType, HexBytes
from eip712.common import SafeTxV1, SafeTxV2, create_safe_tx_def
from eth_typing import HexStr
from eth_utils import add_0x_prefix, to_hex
from pydantic import BaseModel, BeforeValidator, Field, field_validator

from ape_safe.utils import get_safe_tx_hash

SafeTx = Union[SafeTxV1, SafeTxV2]
SafeTxID = NewType("SafeTxID", str)


def clean_api_address(data: Union[AddressType, dict]) -> AddressType:
    # NOTE: Safe API returns `{'value':'<addr>', ...}` object
    if isinstance(data, dict):
        return data["value"]
    return data


def remove_null_params(data):
    if isinstance(data, dict):
        return {
            k: (remove_null_params(v) if isinstance(v, (dict, list, tuple)) else v)
            for k, v in data.items()
            if v is not None
        }

    elif isinstance(data, list):
        return [
            remove_null_params(v) if isinstance(v, (dict, list, tuple)) else v
            for v in data
            if v is not None
        ]

    elif isinstance(data, tuple):
        return tuple(
            remove_null_params(v) if isinstance(v, (dict, list, tuple)) else v
            for v in data
            if v is not None
        )

    else:
        raise ValueError("Can only supply dict, list, or tuple type")


Address = Annotated[AddressType, BeforeValidator(clean_api_address)]


class SafeDetails(BaseModel):
    address: Address
    nonce: int
    threshold: int
    owners: list[Address]
    master_copy: Address = Field(alias="masterCopy")
    modules: list[Address] = []
    fallback_handler: Address = Field(alias="fallbackHandler")
    guard: AddressType
    version: str

    @field_validator("modules", mode="before")
    def convert_none_to_empty_list(cls, value):
        if not value:
            return []
        return value


class SignatureType(str, Enum):
    APPROVED_HASH = "APPROVED_HASH"
    EOA = "EOA"
    ETH_SIGN = "ETH_SIGN"


class SafeTxConfirmation(BaseModel):
    owner: AddressType
    submission_date: datetime = Field(alias="submissionDate")
    transaction_hash: Optional[HexBytes] = Field(default=None, alias="transactionHash")
    signature: HexBytes
    signature_type: Optional[SignatureType] = Field(default=None, alias="signatureType")


class OperationType(int, Enum):
    CALL = 0
    DELEGATECALL = 1


class UnexecutedTxData(BaseModel):
    safe: AddressType
    to: AddressType
    value: int
    data: Optional[HexBytes] = None
    operation: OperationType
    gas_token: AddressType = Field(alias="gasToken")
    safe_tx_gas: int = Field(alias="safeTxGas")
    base_gas: int = Field(alias="baseGas")
    gas_price: int = Field(alias="gasPrice")
    refund_receiver: AddressType = Field(alias="refundReceiver")
    nonce: int
    submission_date: datetime = Field(alias="submissionDate")
    modified: datetime
    safe_tx_hash: SafeTxID = Field(alias="safeTxHash")
    confirmations_required: int = Field(alias="confirmationsRequired")
    confirmations: list[SafeTxConfirmation] = []
    trusted: bool = True
    signatures: Optional[HexBytes] = None

    @classmethod
    def from_safe_tx(cls, safe_tx: SafeTx, confirmations_required: int) -> "UnexecutedTxData":
        return cls(
            safe=safe_tx._verifyingContract_,
            submissionDate=datetime.now(timezone.utc),
            modified=datetime.now(timezone.utc),
            confirmationsRequired=confirmations_required,
            safeTxHash=get_safe_tx_hash(safe_tx),
            **safe_tx._body_["message"],
        )

    def as_safe_tx(self, version: str, chain_id: Optional[int] = None) -> SafeTx:
        tx_def = create_safe_tx_def(
            version=version,
            contract_address=self.safe,
            chain_id=chain_id,
        )

        return tx_def(  # type: ignore[call-arg]
            to=self.to,
            value=self.value,
            data=self.data,
            operation=self.operation,
            gasToken=self.gas_token,
            safeTxGas=self.safe_tx_gas,
            baseGas=self.base_gas,
            gasPrice=self.gas_price,
            refundReceiver=self.refund_receiver,
            nonce=self.nonce,
        )

    @property
    def base_tx_dict(self) -> dict:
        return {
            "to": self.to,
            "value": self.value,
            "data": self.data,
            "operation": self.operation,
            "safeTxGas": self.safe_tx_gas,
            "baseGas": self.base_gas,
            "gasPrice": self.gas_price,
            "gasToken": self.gas_token,
            "refundReceiver": self.refund_receiver,
            "nonce": self.nonce,
        }

    def __str__(self) -> str:
        # TODO: Decode data
        data_hex = to_hex(self.data) if self.data else ""
        if len(data_hex) > 40:
            data_hex = f"{data_hex[:18]}....{data_hex[-18:]}"

        # TODO: Handle MultiSend contract differently
        return f"""Tx ID {self.nonce}
   type: {self.operation._name_}
   from: {self.safe}
     to: {self.to}
  value: {self.value / 1e18} ether
   data: {add_0x_prefix(cast(HexStr, data_hex))}
"""


class ExecutedTxData(UnexecutedTxData):
    execution_date: datetime = Field(alias="executionDate")
    block_number: int = Field(alias="blockNumber")
    transaction_hash: HexBytes = Field(alias="transactionHash")
    executor: AddressType
    is_executed: bool = Field(alias="isExecuted")
    is_successful: bool = Field(alias="isSuccessful")
    eth_gas_price: int = Field(alias="ethGasPrice")
    max_fee_per_gas: Optional[int] = Field(default=None, alias="maxFeePerGas")
    max_priority_fee_per_gas: Optional[int] = Field(default=None, alias="maxPriorityFeePerGas")
    gas_used: int = Field(alias="gasUsed")
    fee: int
    origin: str
    data_decoded: Optional[dict] = Field(default=None, alias="dataDecoded")


SafeApiTxData = Union[ExecutedTxData, UnexecutedTxData]


class DelegateInfo(BaseModel):
    safe: AddressType
    delegate: AddressType
    delegator: AddressType
    label: str = ""



from contextlib import ContextDecorator
from typing import TYPE_CHECKING, Optional

from ape.exceptions import AccountsError, ApeException, ContractLogicError, SignatureError

if TYPE_CHECKING:
    from ape.types import AddressType
    from requests import Response


class ApeSafeException(ApeException):
    pass


class ApeSafeError(ApeSafeException, AccountsError):
    """
    An error to raise in place of AccountsError for the ``ape-safe`` plugin.
    """


class NotASigner(ApeSafeException):
    def __init__(self, signer: "AddressType"):
        super().__init__(f"{signer} is not a valid signer.")


class NoLocalSigners(ApeSafeException, SignatureError):
    def __init__(self):
        super().__init__("No local signers available, try resubmitting with `submitter=` kwarg.")


class NotEnoughSignatures(ApeSafeException, SignatureError):
    def __init__(self, expected: int, actual: int):
        super().__init__(
            f"Not enough signatures, {expected - actual} more are needed. Bypass this behavior"
            " and publish to Safe API by adding 'submit_transaction=False' to your call."
        )


class ClientUnavailable(ApeSafeException):
    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "Client unavailable.")


SAFE_ERROR_CODES = {
    "GS000": "Could not finish initialization",
    "GS001": "Threshold needs to be defined",
    "GS010": "Not enough gas to execute Safe transaction",
    "GS011": "Could not pay gas costs with ether",
    "GS012": "Could not pay gas costs with token",
    "GS013": "Safe transaction failed when gasPrice and safeTxGas were 0",
    "GS020": "Signatures data too short",
    "GS021": "Invalid contract signature location: inside static part",
    "GS022": "Invalid contract signature location: length not present",
    "GS023": "Invalid contract signature location: data not complete",
    "GS024": "Invalid contract signature provided",
    "GS025": "Hash has not been approved",
    "GS026": "Invalid owner provided",
    "GS030": "Only owners can approve a hash",
    "GS031": "Method can only be called from this contract",
    "GS100": "Modules have already been initialized",
    "GS101": "Invalid module address provided",
    "GS102": "Module has already been added",
    "GS103": "Invalid prevModule, module pair provided",
    "GS104": "Method can only be called from an enabled module",
    "GS105": "Invalid starting point for fetching paginated modules",
    "GS106": "Invalid page size for fetching paginated modules",
    "GS200": "Owners have already been set up",
    "GS201": "Threshold cannot exceed owner count",
    "GS202": "Threshold needs to be greater than 0",
    "GS203": "Invalid owner address provided",
    "GS204": "Address is already an owner",
    "GS205": "Invalid prevOwner, owner pair provided",
    "GS300": "Guard does not implement IERC165",
}


class SafeLogicError(ApeSafeException, ContractLogicError):
    def __init__(self, error_code: str):
        super().__init__(f"{SAFE_ERROR_CODES[error_code]} ({error_code})")


class handle_safe_logic_error(ContextDecorator):
    def __enter__(self):
        pass

    def __exit__(self, exc_type: type[BaseException], exc: BaseException, tb):
        if (
            isinstance(exc, ContractLogicError)  # NOTE: Just for mypy
            and exc_type == ContractLogicError
        ):
            message = exc.message.replace("revert: ", "").strip()
            if message.startswith("GS") and message in SAFE_ERROR_CODES:
                raise SafeLogicError(exc.message.replace("revert: ", "")) from exc

        # NOTE: Will raise `exc` by default because we did not return anything


class MulticallException(ApeSafeException):
    pass


class ValueRequired(MulticallException):
    def __init__(self, amount: int):
        super().__init__(f"This transaction must send at least '{amount / 1e18}' ether.")


class UnsupportedChainError(MulticallException):
    def __init__(self):
        super().__init__("Multicall not supported on this chain.")


class SafeClientException(ApeSafeException):
    pass


class ActionNotPerformedError(SafeClientException):
    def __init__(self, message: str):
        super().__init__(message)


class ClientResponseError(SafeClientException):
    def __init__(self, endpoint_url: str, response: "Response", message: Optional[str] = None):
        self.endpoint_url = endpoint_url
        self.response = response
        message = message or f"Exception when calling '{endpoint_url}':\n{response.text}"
        super().__init__(message)


class MultisigTransactionNotFoundError(ClientResponseError):
    def __init__(self, tx_hash: str, endpoint_url: str, response: "Response"):
        message = f"Multisig transaction '{tx_hash}' not found."
        super().__init__(endpoint_url, response, message=message)
